import torch
import os.path as osp

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def ipd_batched(bs, gamma_inner=0.96):
    dims = [5, 5]
    payout_mat_1 = torch.Tensor([[-1, -3], [0, -2]]).to(device)
    payout_mat_2 = payout_mat_1.T
    payout_mat_1 = payout_mat_1.reshape((1, 2, 2)).repeat(bs, 1, 1).to(device)
    payout_mat_2 = payout_mat_2.reshape((1, 2, 2)).repeat(bs, 1, 1).to(device)

    def Ls(th):
        p_1_0 = torch.sigmoid(th[0][:, 0:1])
        p_2_0 = torch.sigmoid(th[1][:, 0:1])
        p = torch.cat(
            [
                p_1_0 * p_2_0,
                p_1_0 * (1 - p_2_0),
                (1 - p_1_0) * p_2_0,
                (1 - p_1_0) * (1 - p_2_0),
            ],
            dim=-1,
        )
        p_1 = torch.reshape(torch.sigmoid(th[0][:, 1:5]), (bs, 4, 1))
        p_2 = torch.reshape(
            torch.sigmoid(
                torch.cat(
                    [th[1][:, 1:2], th[1][:, 3:4], th[1][:, 2:3], th[1][:, 4:5]], dim=-1
                )
            ),
            (bs, 4, 1),
        )
        P = torch.cat(
            [p_1 * p_2, p_1 * (1 - p_2), (1 - p_1) * p_2, (1 - p_1) * (1 - p_2)], dim=-1
        )

        M = torch.matmul(
            p.unsqueeze(1), torch.inverse(torch.eye(4).to(device) - gamma_inner * P)
        )
        L_1 = -torch.matmul(M, torch.reshape(payout_mat_1, (bs, 4, 1)))
        L_2 = -torch.matmul(M, torch.reshape(payout_mat_2, (bs, 4, 1)))
        return [L_1.squeeze(-1), L_2.squeeze(-1), M]

    return dims, Ls


def imp_batched(bs, gamma_inner=0.96):
    dims = [5, 5]
    payout_mat_1 = torch.Tensor([[-1, 1], [1, -1]]).to(device)
    payout_mat_2 = -payout_mat_1
    payout_mat_1 = payout_mat_1.reshape((1, 2, 2)).repeat(bs, 1, 1).to(device)
    payout_mat_2 = payout_mat_2.reshape((1, 2, 2)).repeat(bs, 1, 1).to(device)

    def Ls(th):
        p_1_0 = torch.sigmoid(th[0][:, 0:1])
        p_2_0 = torch.sigmoid(th[1][:, 0:1])
        p = torch.cat(
            [
                p_1_0 * p_2_0,
                p_1_0 * (1 - p_2_0),
                (1 - p_1_0) * p_2_0,
                (1 - p_1_0) * (1 - p_2_0),
            ],
            dim=-1,
        )
        p_1 = torch.reshape(torch.sigmoid(th[0][:, 1:5]), (bs, 4, 1))
        p_2 = torch.reshape(
            torch.sigmoid(
                torch.cat(
                    [th[1][:, 1:2], th[1][:, 3:4], th[1][:, 2:3], th[1][:, 4:5]], dim=-1
                )
            ),
            (bs, 4, 1),
        )
        P = torch.cat(
            [p_1 * p_2, p_1 * (1 - p_2), (1 - p_1) * p_2, (1 - p_1) * (1 - p_2)], dim=-1
        )

        M = torch.matmul(
            p.unsqueeze(1), torch.inverse(torch.eye(4).to(device) - gamma_inner * P)
        )
        L_1 = -torch.matmul(M, torch.reshape(payout_mat_1, (bs, 4, 1)))
        L_2 = -torch.matmul(M, torch.reshape(payout_mat_2, (bs, 4, 1)))
        return [L_1.squeeze(-1), L_2.squeeze(-1), M]

    return dims, Ls


def chicken_game_batch(batch_size=128):
    dims = [1, 1]
    payout_mat_1 = torch.Tensor([[0, -1], [1, -100]]).to(device)
    payout_mat_2 = torch.Tensor([[0, 1], [-1, -100]]).to(device)
    payout_mat_1 = payout_mat_1.reshape((1, 2, 2)).repeat(batch_size, 1, 1)
    payout_mat_2 = payout_mat_2.reshape((1, 2, 2)).repeat(batch_size, 1, 1)

    def Ls(th):
        p_1, p_2 = torch.sigmoid(th[0]), torch.sigmoid(th[1])
        x, y = torch.cat([p_1, 1 - p_1], dim=-1), torch.cat([p_2, 1 - p_2], dim=-1)
        L_1 = -torch.matmul(torch.matmul(x.unsqueeze(1), payout_mat_1), y.unsqueeze(-1))
        L_2 = -torch.matmul(torch.matmul(x.unsqueeze(1), payout_mat_2), y.unsqueeze(-1))
        return [L_1.squeeze(-1), L_2.squeeze(-1), None]

    return dims, Ls


def get_gradient(function, param):
    grad = torch.autograd.grad(function, param, create_graph=True, allow_unused=True)[0]
    return grad


class MetaGames:
    def __init__(self, b, opponent="NL", game="IPD"):
        """
        Opponent can be:
        NL = Naive Learner (gradient updates through environment).
        LOLA = Gradient through NL.
        M-MAML = Loads initial parameters from file.
        STATIC = Doesn't learn.
        """
        self.gamma_inner = 0.96
        self.b = b

        self.game = game
        if self.game == "IPD":
            d, self.game_batched = ipd_batched(b, gamma_inner=self.gamma_inner)
            self.std = 1
            self.lr = 1
        elif self.game == "IMP":
            d, self.game_batched = imp_batched(b, gamma_inner=self.gamma_inner)
            self.std = 1
            self.lr = 0.1
        elif self.game == "chicken":
            d, self.game_batched = chicken_game_batch(b)
            self.std = 1
            self.lr = 1
        else:
            raise NotImplementedError

        self.d = d[0]

        self.opponent = opponent
        if self.opponent == "MAMAML":
            f = f"m-maml_{self.game}.th"
            assert osp.exists(f), "Generate the M-MAML weights first"
            self.init_th_ba = torch.load(f)
        else:
            self.init_th_ba = None

    def reset(self, info=False):
        if self.init_th_ba is not None:
            self.inner_th_ba = self.init_th_ba.detach() * torch.ones(
                (self.b, self.d), requires_grad=True
            ).to(device)
        else:
            self.inner_th_ba = torch.nn.init.normal_(
                torch.empty((self.b, self.d), requires_grad=True), std=self.std
            ).to(device)
        outer_th_ba = torch.nn.init.normal_(
            torch.empty((self.b, self.d), requires_grad=True), std=self.std
        ).to(device)
        state, _, _, M = self.step(outer_th_ba)
        if info:
            return state, M
        else:
            return state

    def step(self, outer_th_ba):
        last_inner_th_ba = self.inner_th_ba.detach().clone()
        if self.opponent == "NL" or self.opponent == "M-MAML":
            th_ba = [self.inner_th_ba, outer_th_ba.detach()]
            l1, l2, M = self.game_batched(th_ba)
            grad = get_gradient(l1.sum(), self.inner_th_ba)
            with torch.no_grad():
                self.inner_th_ba -= grad * self.lr
        elif self.opponent == "LOLA":
            th_ba = [self.inner_th_ba, outer_th_ba.detach()]
            th_ba[1].requires_grad = True
            l1, l2, M = self.game_batched(th_ba)
            losses = [l1, l2]
            grad_L = [
                [get_gradient(losses[j].sum(), th_ba[i]) for j in range(2)]
                for i in range(2)
            ]
            term = (grad_L[1][0] * grad_L[1][1]).sum()
            grad = grad_L[0][0] - self.lr * get_gradient(term, th_ba[0])
            with torch.no_grad():
                self.inner_th_ba -= grad * self.lr
        elif self.opponent == "STATIC":
            th_ba = [self.inner_th_ba, outer_th_ba.detach()]
            l1, l2, M = self.game_batched(th_ba)
        else:
            raise NotImplementedError

        if self.game == "IPD" or self.game == "IMP":
            return (
                torch.sigmoid(
                    torch.cat((outer_th_ba, last_inner_th_ba), dim=-1)
                ).detach(),
                (-l2 * (1 - self.gamma_inner)).detach(),
                (-l1 * (1 - self.gamma_inner)).detach(),
                M,
            )
        else:
            return (
                torch.sigmoid(
                    torch.cat((outer_th_ba, last_inner_th_ba), dim=-1)
                ).detach(),
                -l2.detach(),
                -l1.detach(),
                M,
            )


class SymmetricMetaGames:
    def __init__(self, b, game="IPD"):
        self.gamma_inner = 0.96

        self.b = b
        self.game = game
        if self.game == "IPD":
            d, self.game_batched = ipd_batched(b, gamma_inner=self.gamma_inner)
            self.std = 1
        elif self.game == "IMP":
            d, self.game_batched = imp_batched(b, gamma_inner=self.gamma_inner)
            self.std = 1
        elif self.game == "chicken":
            d, self.game_batched = chicken_game_batch(b)
            self.std = 1
        else:
            raise NotImplementedError

        self.d = d[0]

    def reset(self, info=False):
        p_ba_0 = torch.nn.init.normal_(torch.empty((self.b, self.d)), std=self.std).to(
            device
        )
        p_ba_1 = torch.nn.init.normal_(torch.empty((self.b, self.d)), std=self.std).to(
            device
        )
        state_0 = torch.sigmoid(torch.cat((p_ba_0.detach(), p_ba_1.detach()), dim=-1))
        state_1 = torch.sigmoid(torch.cat((p_ba_1.detach(), p_ba_0.detach()), dim=-1))

        if info:
            state, _, M = self.step(p_ba_0, p_ba_1)
            return state, M
        else:
            return [state_0, state_1]

    def step(self, p_ba_0, p_ba_1):
        th_ba = [p_ba_0.detach(), p_ba_1.detach()]
        l1, l2, M = self.game_batched(th_ba)
        state_0 = torch.sigmoid(torch.cat((p_ba_0.detach(), p_ba_1.detach()), dim=-1))
        state_1 = torch.sigmoid(torch.cat((p_ba_1.detach(), p_ba_0.detach()), dim=-1))
        if self.game == "IPD" or self.game == "IMP":
            return (
                [state_0, state_1],
                [-l1 * (1 - self.gamma_inner), -l2 * (1 - self.gamma_inner)],
                M,
            )
        else:
            return [state_0, state_1], [-l1.detach(), -l2.detach()], M


class NonMfosMetaGames:
    def __init__(self, b, p1="NL", p2="NL", game="IPD", mmapg_id=None):
        """
        Opponent can be:
        NL = Naive Learner (gradient updates through environment).
        LOLA = Gradient through NL.
        STATIC = Doesn't learn.
        M-MAML = Loads initial parameters from file.
        """
        self.gamma_inner = 0.96
        self.b = b

        self.p1 = p1
        self.p2 = p2
        self.game = game
        if self.game == "IPD":
            d, self.game_batched = ipd_batched(b, gamma_inner=self.gamma_inner)
            self.std = 1
            self.lr = 1
        elif self.game == "IMP":
            d, self.game_batched = imp_batched(b, gamma_inner=self.gamma_inner)
            self.std = 1
            self.lr = 0.1
        elif self.game == "chicken":
            d, self.game_batched = chicken_game_batch(b)
            self.std = 1
            self.lr = 1
        else:
            raise NotImplementedError

        self.d = d[0]

        self.init_th_ba = None
        if self.p1 == "M-MAML" or self.p2 == "M-MAML":
            f = f"m-maml_{self.game}.th"
            assert osp.exists(f), "Generate the MAMAML weights first"
            self.init_th_ba = torch.load(f)

    def reset(self, info=False):
        self.p1_th_ba = torch.nn.init.normal_(
            torch.empty((self.b, self.d), requires_grad=True), std=self.std
        ).to(device)
        self.p2_th_ba = torch.nn.init.normal_(
            torch.empty((self.b, self.d), requires_grad=True), std=self.std
        ).to(device)

        if self.p1 == "M-MAML":
            self.p1_th_ba = self.init_th_ba.detach() * torch.ones(
                (self.b, self.d), requires_grad=True
            ).to(device)
        if self.p2 == "M-MAML":
            self.p2_th_ba = self.init_th_ba.detach() * torch.ones(
                (self.b, self.d), requires_grad=True
            ).to(device)

        state, _, _, M = self.step()
        if info:
            return state, M

        return None

    def step(self, info=False):
        last_p1_th_ba = self.p1_th_ba.clone()
        last_p2_th_ba = self.p2_th_ba.clone()
        th_ba = [self.p2_th_ba, self.p1_th_ba]
        l1, l2, M = self.game_batched(th_ba)

        # UPDATE P1
        if self.p1 == "NL" or self.p1 == "M-MAML":
            grad = get_gradient(l2.sum(), self.p1_th_ba)
            with torch.no_grad():
                self.p1_th_ba -= grad * self.lr
        elif self.p1 == "LOLA":
            losses = [l1, l2]
            grad_L = [
                [get_gradient(losses[j].sum(), th_ba[i]) for j in range(2)]
                for i in range(2)
            ]
            term = (grad_L[0][0] * grad_L[0][1]).sum()
            grad = grad_L[1][1] - self.lr * get_gradient(term, th_ba[1])
            with torch.no_grad():
                self.p1_th_ba -= grad * self.lr
        elif self.p1 == "STATIC":
            pass
        else:
            raise NotImplementedError

        # UPDATE P2
        if self.p2 == "NL" or self.p2 == "M-MAML":
            grad = get_gradient(l1.sum(), self.p2_th_ba)
            with torch.no_grad():
                self.p2_th_ba -= grad * self.lr
        elif self.p2 == "LOLA":
            losses = [l1, l2]
            grad_L = [
                [get_gradient(losses[j].sum(), th_ba[i]) for j in range(2)]
                for i in range(2)
            ]
            term = (grad_L[1][0] * grad_L[1][1]).sum()
            grad = grad_L[0][0] - self.lr * get_gradient(term, th_ba[0])
            with torch.no_grad():
                self.p2_th_ba -= grad * self.lr
        elif self.p2 == "STATIC":
            pass
        else:
            raise NotImplementedError

        if self.game == "IPD" or self.game == "IMP":
            return (
                torch.sigmoid(torch.cat([last_p1_th_ba, last_p2_th_ba])),
                -l2 * (1 - self.gamma_inner),
                -l1 * (1 - self.gamma_inner),
                M,
            )
        else:
            return (
                torch.sigmoid(torch.cat([last_p1_th_ba, last_p2_th_ba])),
                -l2.detach(),
                -l1.detach(),
                M,
            )
